@@ -1,24 +1,27 @@
 # coding: utf-8
 """
-Flask-OAuthlib
---------------
+    flask_oauthlib.client
+    ~~~~~~~~~~~~~~~~~~~~~
 
-Implemnts OAuth1 and OAuth2 support for Flask.
+    Implemnts OAuth1 and OAuth2 support for Flask.
 
-:copyright: (c) 2013 by Hsiaoming Yang.
+    :copyright: (c) 2013 by Hsiaoming Yang.
 """
 
 import urllib2
+import logging
 import oauthlib.oauth1
 import oauthlib.oauth2
 from functools import wraps
 from oauthlib.common import to_unicode
-from urlparse import urljoin
+from urlparse import urljoin, urlparse
 from flask import request, redirect, json, session
 from werkzeug import url_quote, url_decode, url_encode, parse_options_header
 
 
-__all__ = ['OAuth', 'OAuthRemoteApp', 'OAuthResponse', 'OAuthException']
+__all__ = ('OAuth', 'OAuthRemoteApp', 'OAuthResponse', 'OAuthException')
+
+log = logging.getLogger('flask_oauthlib')
 
 
 class OAuth(object):
@@ -34,6 +37,7 @@ class OAuth(object):
     def __init__(self, app=None):
         self.remote_apps = {}
 
+        self.app = app
         if app:
             self.init_app(app)
 
@@ -47,7 +51,7 @@ class OAuth(object):
         """
         self.app = app
         app.extensions = getattr(app, 'extensions', {})
-        app.extensions['oauth-client'] = self
+        app.extensions['oauthlib.client'] = self
 
     def remote_app(self, name, register=True, **kwargs):
         """Registers a new remote application.
@@ -58,11 +62,14 @@ class OAuth(object):
         Find more parameters from :class:`OAuthRemoteApp`.
         """
 
-        app = OAuthRemoteApp(self, name, **kwargs)
+        if self.app and self.app.testing:
+            kwargs['test_client'] = self.app.test_client()
+
+        remote = OAuthRemoteApp(self, name, **kwargs)
         if register:
             assert name not in self.remote_apps
-            self.remote_apps[name] = app
-        return app
+            self.remote_apps[name] = remote
+        return remote
 
     def __getattr__(self, key):
         try:
@@ -77,7 +84,7 @@ class OAuth(object):
 _etree = None
 
 
-def get_etree():
+def get_etree():  # pragma: no cover
     global _etree
     if _etree is not None:
         return _etree
@@ -115,7 +122,8 @@ def parse_response(resp, content, strict=False, content_type=None):
     return url_decode(content, charset=charset).to_dict()
 
 
-def make_request(uri, headers=None, data=None, method=None):
+def make_request(uri, headers=None, data=None, method=None,
+                 test_client=None):
     if headers is None:
         headers = {}
 
@@ -128,6 +136,19 @@ def make_request(uri, headers=None, data=None, method=None):
         uri = add_query(uri, data)
         data = None
 
+    log.debug('Request %r with %r method' % (uri, method))
+
+    if test_client:
+        # test client is a `werkzeug.test.Client`
+        parsed = urlparse(uri)
+        uri = '%s?%s' % (parsed.path, parsed.query)
+        resp = test_client.open(
+            uri, headers=headers, data=data, method=method
+        )
+        # for compatible
+        resp.code = resp.status_code
+        return resp, resp.data
+
     req = urllib2.Request(uri, headers=headers, data=data)
     req.get_method = lambda: method.upper()
     try:
@@ -215,6 +236,7 @@ class OAuthRemoteApp(object):
         access_token_params=None,
         access_token_method='GET',
         content_type=None,
+        test_client=None,
         encoding='utf-8',
     ):
 
@@ -226,13 +248,15 @@ class OAuthRemoteApp(object):
         self.authorize_url = authorize_url
         self.consumer_key = consumer_key
         self.consumer_secret = consumer_secret
-        self.tokengetter_func = None
         self.request_token_params = request_token_params or {}
         self.access_token_params = access_token_params or {}
         self.access_token_method = access_token_method
         self.content_type = content_type
+        self.test_client = test_client
         self.encoding = encoding
 
+        self._tokengetter = None
+
     def make_client(self, token=None):
         # request_token_url is for oauth1
         if self.request_token_url:
@@ -329,7 +353,8 @@ class OAuthRemoteApp(object):
             uri, headers, body = self.pre_request(uri, headers, body)
 
         resp, content = make_request(
-            uri, headers, data=body, method=method
+            uri, headers, data=body, method=method,
+            test_client=self.test_client
         )
         return OAuthResponse(resp, content, self.content_type)
 
@@ -367,7 +392,7 @@ class OAuthRemoteApp(object):
         """
         Register a function as token getter.
         """
-        self.tokengetter_func = f
+        self._tokengetter = f
         return f
 
     def expand_url(self, url):
@@ -383,7 +408,9 @@ class OAuthRemoteApp(object):
         uri, headers, _ = client.sign(
             self.expand_url(self.request_token_url)
         )
-        resp, content = make_request(uri, headers)
+        resp, content = make_request(
+            uri, headers, test_client=self.test_client
+        )
         if resp.code not in (200, 201):
             raise OAuthException(
                 'Failed to generate request token',
@@ -400,8 +427,8 @@ class OAuthRemoteApp(object):
         return tup
 
     def get_request_token(self):
-        assert self.tokengetter_func is not None, 'missing tokengetter'
-        rv = self.tokengetter_func()
+        assert self._tokengetter is not None, 'missing tokengetter'
+        rv = self._tokengetter()
         if rv is None:
             raise OAuthException('No token available', type='token_missing')
         return rv
@@ -419,7 +446,9 @@ class OAuthRemoteApp(object):
             _encode(self.access_token_method)
         )
 
-        resp, content = make_request(uri, headers, data)
+        resp, content = make_request(
+            uri, headers, data, test_client=self.test_client
+        )
         data = parse_response(resp, content)
         if resp.code not in (200, 201):
             raise OAuthException(
@@ -437,22 +466,28 @@ class OAuthRemoteApp(object):
             'client_secret': self.consumer_secret,
             'redirect_uri': session.get('%s_oauthredir' % self.name)
         }
+        log.debug('Prepare oauth2 remote args %r', remote_args)
         remote_args.update(self.access_token_params)
         if self.access_token_method == 'POST':
             body = client.prepare_request_body(**remote_args)
             resp, content = make_request(
                 self.expand_url(self.access_token_url),
                 data=body,
-                method=self.access_token_method
+                method=self.access_token_method,
+                test_client=self.test_client,
             )
         elif self.access_token_method == 'GET':
-            url = client.prepare_request_uri(
-                self.expand_url(self.access_token_url), **remote_args
+            qs = client.prepare_request_body(**remote_args)
+            url = self.expand_url(self.access_token_url)
+            url += ('?' in url and '&' or '?') + qs
+            resp, content = make_request(
+                url,
+                method=self.access_token_method,
+                test_client=self.test_client,
             )
-            resp, content = make_request(url, method=self.access_token_method)
         else:
             raise OAuthException(
-                'Unsupported access_token_method: %s' % \
+                'Unsupported access_token_method: ' %
                 self.access_token_method
             )
 
@@ -464,13 +499,23 @@ class OAuthRemoteApp(object):
             )
         return data
 
+    def handle_unknown_response(self):
+        """Handles a unknown authorization response."""
+        return None
+
     def authorized_handler(self, f):
         @wraps(f)
         def decorated(*args, **kwargs):
             if 'oauth_verifier' in request.args:
-                data = self.handle_oauth1_response()
+                try:
+                    data = self.handle_oauth1_response()
+                except OAuthException as e:
+                    data = e
             elif 'code' in request.args:
-                data = self.handle_oauth2_response()
+                try:
+                    data = self.handle_oauth2_response()
+                except OAuthException as e:
+                    data = e
             else:
                 data = self.handle_unknown_response()
 
